from .models import *
import unittest


class PresentationTestCase(unittest.TestCase):

    def setUp(self):
        nrm.lua.drop(args=['*'])

    def test_fields(self):
        office = Office(
            name = 'Fleety',
            address = 'Springfield 132',
        ).save()

        self.assertDictEqual(office.to_json(), {
            'id': office.id,
            'name': office.name,
            'address': office.address,
            '_type': 'office',
        })

        self.assertDictEqual(office.to_json(fields=['name']), {
            'name': 'Fleety',
        })

        self.assertDictEqual(office.to_json(fields=['id', 'name']), {
            'id': office.id,
            'name': office.name,
        })

        self.assertDictEqual(office.to_json(fields=['none']), {})

    def test_embed(self):
        office = Office(
            name = 'Fleety',
            address = 'Springfield 132',
        ).save()
        employee = Employee(
            name = 'Juan',
            last_name = 'Alduci',
        ).save()
        office.proxy.employees.set([employee])

        self.assertDictEqual(office.to_json(embed=['employees']), {
            'id': office.id,
            'name': office.name,
            'address': office.address,
            'employees': [employee.to_json()],
            '_type': 'office',
        })

        self.assertDictEqual(office.to_json(embed=['employees.name']), {
            'id': office.id,
            'name': office.name,
            'address': office.address,
            'employees': [employee.to_json(fields=['name'])],
            '_type': 'office',
        })

        self.assertDictEqual(office.to_json(embed=['employees.name', 'employees.id']), {
            'id': office.id,
            'name': office.name,
            'address': office.address,
            'employees': [employee.to_json(fields=['name', 'id'])],
            '_type': 'office',
        })

    def test_embed_foreignid_relation(self):
        office = Office(
            name = 'Fleety',
            address = 'Springfield 132',
        ).save()
        employee = Employee(
            name = 'Juan',
            last_name = 'Alduci',
        ).save()
        office.proxy.employees.set([employee])

        self.assertDictEqual(employee.to_json(embed=['office']), {
            'id': employee.id,
            'name': employee.name,
            'last_name': employee.last_name,
            '_type': 'employee',
            'office': office.to_json(),
        })

        self.assertDictEqual(employee.to_json(embed=['office.name']), {
            'id': employee.id,
            'name': employee.name,
            'last_name': employee.last_name,
            '_type': 'employee',
            'office': office.to_json(fields=['name']),
        })

        self.assertDictEqual(employee.to_json(embed=['office.name', 'office.id']), {
            'id': employee.id,
            'name': employee.name,
            'last_name': employee.last_name,
            '_type': 'employee',
            'office': office.to_json(fields=['name', 'id']),
        })

        office.proxy.employees.remove(employee)

        self.assertDictEqual(employee.to_json(embed=['office']), {
            'id': employee.id,
            'name': employee.name,
            'last_name': employee.last_name,
            '_type': 'employee',
            'office': None,
        })

        self.assertDictEqual(employee.to_json(embed=['office.name']), {
            'id': employee.id,
            'name': employee.name,
            'last_name': employee.last_name,
            '_type': 'employee',
            'office': None,
        })

        self.assertDictEqual(employee.to_json(embed=['office.name', 'office.id']), {
            'id': employee.id,
            'name': employee.name,
            'last_name': employee.last_name,
            '_type': 'employee',
            'office': None,
        })



from coralillo import Model, Engine
from coralillo.fields import Text, SetRelation, ForeignIdRelation

nrm = Engine()

class TestModel(Model):
    class Meta:
        engine = nrm

# For one-many relations
class Pet(TestModel):
    name = Text()
    owner = ForeignIdRelation('coralillo.tests.models.Person', inverse='pets')


class Person(TestModel):
    name = Text()
    pets = SetRelation(Pet, on_delete='cascade', inverse='owner')


# For many to many relations
class Driver(TestModel):
    name = Text()
    cars = SetRelation('coralillo.tests.models.Car', inverse='drivers')


class Car(TestModel):
    name = Text()
    drivers = SetRelation(Driver, inverse='cars')


class Employee(TestModel):
    name = Text()
    last_name = Text()
    office = ForeignIdRelation('coralillo.tests.models.Office', inverse='employees')


class Office(TestModel):
    name = Text()
    address = Text()
    employees = SetRelation(Employee, inverse='office')



from coralillo import Model, BoundedModel, fields, Engine
from coralillo.auth import PermissionHolder
import unittest

nrm = Engine()


class User(Model, PermissionHolder):
    name = fields.Text()

    class Meta:
        engine = nrm

class Pet(BoundedModel):
    name = fields.Text()

    class Meta:
        engine = nrm

    @classmethod
    def prefix(cls):
        return 'bound'


class PermissionTestCase(unittest.TestCase):

    def setUp(self):
        nrm.lua.drop(args=['*'])
        self.user = User(
            name      = 'juan',
        ).save()
        self.allow_key = self.user.allow_key()

    def test_allow_key(self):
        self.user.allow('a')

        self.assertTrue(nrm.redis.exists('user:{}:allow'.format(self.user.id)))

    def test_add_permissions(self):
        self.user.allow('a:b:c')
        self.assertTrue(nrm.redis.sismember(self.allow_key, 'a:b:c'))

    def test_add_ignores_when_has_parent(self):
        self.user.allow('a')
        self.user.allow('a:b')
        self.user.allow('a:b:c')

        self.assertTrue(nrm.redis.sismember(self.allow_key, 'a'))
        self.assertFalse(nrm.redis.sismember(self.allow_key, 'a:b'))
        self.assertFalse(nrm.redis.sismember(self.allow_key, 'a:b:c'))

        self.user.allow('foo:var')
        self.user.allow('foo:var:log')
        self.assertFalse(nrm.redis.sismember(self.allow_key, 'foo'))
        self.assertTrue(nrm.redis.sismember(self.allow_key, 'foo:var'))
        self.assertFalse(nrm.redis.sismember(self.allow_key, 'foo:var:log'))

    def test_add_deletes_lower(self):
        self.user.allow('a:b/v')
        self.user.allow('a/v')

        self.assertSetEqual(self.user.get_perms(), set(['a/v']))

    def test_revoke_permission(self):
        self.user.allow('a:b')
        self.user.revoke('a:b')

        self.assertFalse(nrm.redis.sismember(self.allow_key, 'a:b'))

    def test_check_permission_inheritance(self):
        self.user.allow('a:b')

        self.assertTrue(self.user.is_allowed('a:b'))
        self.assertTrue(self.user.is_allowed('a:b:c'))

        self.assertFalse(self.user.is_allowed('a'))
        self.assertFalse(self.user.is_allowed('a:d'))

    def test_can_carry_restrict(self):
        self.user.allow('org:fleet/view')

        self.assertTrue(self.user.is_allowed('org:fleet:somefleet/view'))

    def test_permission_key(self):
        self.assertEqual(self.user.permission(), 'user:{}'.format(self.user.id))
        self.assertEqual(self.user.permission(restrict='view'), 'user:{}/view'.format(self.user.id))

    def test_minor_ignored_if_mayor(self):
        self.user.allow('org:fleet/view')
        self.user.allow('org:fleet:325234/view')

        self.assertSetEqual(self.user.get_perms(), set(['org:fleet/view']))

    def test_perm_framework_needs_strings(self):
        with self.assertRaises(AssertionError):
            self.user.allow(User)

        with self.assertRaises(AssertionError):
            self.user.is_allowed(User)

        with self.assertRaises(AssertionError):
            self.user.revoke(User)

    def test_permission_function(self):
        self.assertEqual(self.user.permission(), 'user:{}'.format(self.user.id))
        self.assertEqual(self.user.permission('eat'), 'user:{}/eat'.format(self.user.id))

        pet = Pet(name='doggo').save()
        self.assertEqual(pet.permission(), 'bound:pet:{}'.format(pet.id))
        self.assertEqual(pet.permission('walk'), 'bound:pet:{}/walk'.format(pet.id))

    def test_delete_user_deletes_permission_bag(self):
        self.user.allow('foo')

        self.assertTrue(nrm.redis.exists('user:{}:allow'.format(self.user.id)))

        self.user.delete()
        self.assertFalse(nrm.redis.exists('user:{}:allow'.format(self.user.id)))


if __name__ == '__main__':
    unittest.main()



from coralillo import datamodel, Engine, Model, BoundedModel, fields
from coralillo.datamodel import debyte_string
from coralillo.errors import ModelNotFoundError
import unittest

nrm = Engine()

class TestModel(Model):
    class Meta:
        engine = nrm

class Person(TestModel):
    name = fields.Text()


class Ship(TestModel):
    name = fields.Text()
    code = fields.Text(index=True)


class Pet(BoundedModel):
    name = fields.Text()

    @classmethod
    def prefix(cls):
        return 'testing'

    class Meta:
        engine = nrm


class SideWalk(TestModel):
    name = fields.Text()


class House(TestModel):
    number = fields.Integer(required=False)


class ModelTestCase(unittest.TestCase):

    def setUp(self):
        nrm.lua.drop(args=['*'])

    def test_create_user(self):
        user = Person(
            name      = 'John',
        ).save()

        self.assertEqual(user.name, 'John')

        self.assertTrue(nrm.redis.sismember('person:members', user.id))
        self.assertEqual(nrm.redis.hget('person:{}:obj'.format(user.id), 'name'), b'John')

    def test_retrieve_user_by_id(self):
        carla = Person( name = 'Carla',).save()
        roberta = Person( name = 'Roberta',).save()

        read_user = Person.get(carla.id)

        self.assertEqual(read_user.name, carla.name)

    def test_retrieve_by_index(self):
        titan = Ship(name='te trece', code = 'T13',).save()
        atlan = Ship(name='te catorce', code = 'T14',).save()

        found_ship = Ship.get_by('code', 'T13')

        self.assertIsNotNone(found_ship)
        self.assertTrue(titan == found_ship)

    def test_update_keep_index(self):
        ship = Ship(name='the ship', code='TS').save()

        ship.update(name='updated name')

        self.assertEqual(ship.code, 'TS')
        self.assertEqual(ship.name, 'updated name')

        self.assertEqual(debyte_string(nrm.redis.hget('ship:index_code', 'TS')), ship.id)

    def test_update_changes_index(self):
        ship = Ship(code='THECODE').save()

        ship.update(code='NEWCODE')

        self.assertEqual(ship.code, 'NEWCODE')

        self.assertEqual(debyte_string(nrm.redis.hget('ship:index_code', 'NEWCODE')), ship.id)
        self.assertIsNone(nrm.redis.hget('ship:index_code', 'THECODE'))

    def test_get(self):
        org = Person(name='Juan').save()
        got = Person.get(org.id)

        self.assertTrue(org == got)

    def test_get_all(self):
        p1 = Person(name='Juan').save()
        p2 = Person(name='Pepe').save()

        allitems = Person.get_all()

        allitems.sort(key=lambda x: x.name)

        item1 = allitems[0]
        item2 = allitems[1]

        self.assertEqual(item1, p1)
        self.assertEqual(item2, p2)

    def test_bounded_model(self):
        dev = Pet(
            name = 'foo',
        ).save()

        self.assertFalse(nrm.redis.exists('pet:'+dev.id))
        self.assertTrue(nrm.redis.exists('testing:pet:{}:obj'.format(dev.id)))
        self.assertEqual(nrm.redis.hget('testing:pet:{}:obj'.format(dev.id), 'name'), b'foo')
        self.assertTrue(nrm.redis.sismember('testing:pet:members', dev.id))

    def test_delete(self):
        dev = Pet(
            code = 'foo',
        ).save()

        self.assertIsNotNone(Pet.get(dev.id))

        dev.delete()

        self.assertIsNone(Pet.get(dev.id))
        self.assertFalse(nrm.redis.sismember('testing:pet:members', dev.id))

    def test_delete_index(self):
        ship = Ship(code='A12').save()

        self.assertEqual(debyte_string(nrm.redis.hget('ship:index_code', 'A12')), ship.id)

        ship.delete()

        self.assertFalse(nrm.redis.hexists('ship:index_code', 'A12'))

    def test_is_object_key(self):
        ship = Ship(code='A12').save()

        self.assertTrue(Ship.is_object_key(ship.key()))

    def test_fqn(self):
        ship = Ship(code='A12').save()

        self.assertEqual(ship.fqn(), 'ship:{}'.format(ship.id))

    def test_model_table_conversion(self):
        sw = SideWalk(name='foo').save()

        self.assertTrue(nrm.redis.exists('side_walk:members'))
        self.assertTrue(nrm.redis.exists('side_walk:{}:obj'.format(sw.id)))

    def test_object_count(self):
        sw1 = SideWalk(name='1').save()
        self.assertEqual(SideWalk.count(), 1)

        sw2 = SideWalk(name='2').save()
        self.assertEqual(SideWalk.count(), 2)

        sw1.delete()
        self.assertEqual(SideWalk.count(), 1)

    def test_get_or_exception(self):
        with self.assertRaises(ModelNotFoundError):
            SideWalk.get_or_exception('nonsense')

        with self.assertRaises(ModelNotFoundError):
            Ship.get_by_or_exception('code', 'nonsense')

    def test_recover_none_int(self):
        h = House.validate()
        h.save()

        h_rec = House.get(h.id)
        self.assertIsNone(h.number)


if __name__ == '__main__':
    unittest.main()



import unittest
from coralillo import Engine, Model, fields
import json

nrm = Engine()


class Something(Model):
    name = fields.Text()
    notify = True

    class Meta:
        engine = nrm


class EventTestCase(unittest.TestCase):

    def setUp(self):
        nrm.lua.drop(args=['*'])
        self.maxDiff = None

    def test_create_sends_message(self):
        p = nrm.redis.pubsub(ignore_subscribe_messages=True)
        p.psubscribe('something', 'something:*')

        thing = Something(name='the thing', abbr='TFC').save()

        messages = p.listen()

        message = next(messages)
        self.assertIsNotNone(message)
        self.assertEqual(message['channel'], b'something')
        self.assertDictEqual(json.loads(message['data'].decode('utf8')), {
            'event': 'create',
            'data': thing.to_json(),
        })

        message = next(messages)
        self.assertIsNotNone(message)
        self.assertEqual(message['channel'], thing.key().encode())
        self.assertDictEqual(json.loads(message['data'].decode('utf8')), {
            'event': 'create',
            'data': thing.to_json(),
        })

        p.unsubscribe()

    def test_delete_sends_message(self):
        thing = Something(name='the thing', abbr='TFC').save()

        p = nrm.redis.pubsub(ignore_subscribe_messages=True)
        p.psubscribe('something', 'something:*')

        thing.delete()

        messages = p.listen()

        message = next(messages)
        self.assertIsNotNone(message)
        self.assertEqual(message['channel'], b'something')
        self.assertDictEqual(json.loads(message['data'].decode('utf8')), {
            'event': 'delete',
            'data': thing.to_json(),
        })

        message = next(messages)
        self.assertIsNotNone(message)
        self.assertEqual(message['channel'], thing.key().encode())
        self.assertDictEqual(json.loads(message['data'].decode('utf8')), {
            'event': 'delete',
            'data': thing.to_json(),
        })

        p.unsubscribe()

    def test_update_sends_message(self):
        thing = Something(name='the thing', abbr='TFC').save()

        p = nrm.redis.pubsub(ignore_subscribe_messages=True)
        p.psubscribe('something', 'something:*')

        thing.update(name='renamed thing')

        messages = p.listen()

        message = next(messages)
        self.assertIsNotNone(message)
        self.assertEqual(message['channel'], b'something')
        data = json.loads(message['data'].decode('utf8'))
        self.assertDictEqual(data, {
            'event': 'update',
            'data': thing.to_json(),
        })

        message = next(messages)
        self.assertIsNotNone(message)
        self.assertEqual(message['channel'], thing.key().encode())
        data = json.loads(message['data'].decode('utf8'))
        self.assertDictEqual(data, {
            'event': 'update',
            'data': thing.to_json(),
        })

        self.assertEqual(data['data']['name'], 'renamed thing')

        p.unsubscribe()


if __name__ == '__main__':
    unittest.main()



from coralillo.tests.test_engine import EngineTestCase
from coralillo.tests.test_events import EventTestCase
from coralillo.tests.test_fields import FieldTestCase
from coralillo.tests.test_models import ModelTestCase
from coralillo.tests.test_permissions import PermissionTestCase
from coralillo.tests.test_relations import RelationTestCase
from coralillo.tests.test_presentation import PresentationTestCase
from coralillo.tests.test_utils import UtilsTestCase
import unittest


if __name__ == '__main__':
    unittest.main()



from .models import Pet, Person, Driver, Car, nrm
import unittest


class RelationTestCase(unittest.TestCase):

    def setUp(self):
        nrm.lua.drop(args=['*'])

    def test_relation(self):
        org = Car(
            name      = 'Testing Inc',
        ).save()

        user = Driver(
            name      = 'Sam',
        ).save()
        user.proxy.cars.set([org])

        self.assertEqual(user.proxy.cars.count(), 1)

        self.assertEqual(user.cars[0].id, org.id)
        self.assertTrue(nrm.redis.sismember('driver:{}:srel_cars'.format(user.id), org.id))

        same = Driver.get(user.id)
        same.proxy.cars.fill()

        self.assertEqual(same.cars[0].id, org.id)
        self.assertEqual(same.to_json(embed=['cars'])['cars'][0]['name'], 'Testing Inc')

    def test_delete_with_related(self):
        doggo = Pet(name='doggo').save()
        catto = Pet(name='catto').save()

        owner = Person(
            name = 'John',
        ).save()
        owner.proxy.pets.set([doggo, catto])

        self.assertTrue(doggo in owner.proxy.pets)
        self.assertTrue(catto in owner.proxy.pets)

        owner.delete()

        self.assertIsNone(Pet.get(doggo.id))
        self.assertIsNone(Pet.get(catto.id))

    def test_many_to_many_relationship(self):
        u1 = Driver(name='u1').save()
        u2 = Driver(name='u2').save()
        o1 = Car(name='o1').save()
        o2 = Car(name='o2').save()

        # test adding a relationship creates the inverse
        u1.proxy.cars.set([o1, o2])

        o1.proxy.drivers.fill()
        o2.proxy.drivers.fill()

        self.assertTrue(type(o1.drivers) == list)
        self.assertTrue(type(o2.drivers) == list)

        self.assertEqual(len(o1.drivers), 1)
        self.assertEqual(len(o2.drivers), 1)

        self.assertEqual(o1.drivers[0].name, 'u1')
        self.assertEqual(o2.drivers[0].name, 'u1')

        u2.proxy.cars.set([o2])

        o1.proxy.drivers.fill()
        o2.proxy.drivers.fill()

        self.assertTrue(type(o1.drivers) == list)
        self.assertTrue(type(o2.drivers) == list)

        self.assertEqual(len(o1.drivers), 1)
        self.assertEqual(len(o2.drivers), 2)

        o2.drivers.sort(key=lambda x:x.name)

        self.assertEqual(o1.drivers[0].name, 'u1')
        self.assertEqual(o2.drivers[0].name, 'u1')
        self.assertEqual(o2.drivers[1].name, 'u2')

        # test deleting an object deletes the relationship in the related

        u1.delete()

        o1.proxy.drivers.fill()
        o2.proxy.drivers.fill()

        self.assertTrue(type(o1.drivers) == list)
        self.assertTrue(type(o2.drivers) == list)

        self.assertEqual(len(o1.drivers), 0)
        self.assertEqual(len(o2.drivers), 1)

    def test_querying_related(self):
        o1 = Car(name='o1').save()
        u1 = Driver(name='u1').save()
        u2 = Driver(name='u2').save()
        u1.proxy.cars.set([o1])

        self.assertTrue(u1 in o1.proxy.drivers)
        self.assertFalse(u2 in o1.proxy.drivers)

    def test_foreign_key(self):
        owner = Person(name='John').save()
        pet = Pet(name='doggo').save()

        pet.proxy.owner.set(owner)
        owner.proxy.pets.fill()

        self.assertIsNotNone(pet.owner)
        self.assertEqual(pet.owner.id, owner.id)

        self.assertTrue(type(owner.pets) == list)
        self.assertEqual(len(owner.pets), 1)
        self.assertEqual(owner.pets[0].id, pet.id)

        pet.delete()

        pet.proxy.owner.fill()
        owner.proxy.pets.fill()

        self.assertTrue(type(owner.pets) == list)
        self.assertEqual(len(owner.pets), 0)

    def test_foreign_key_inverse(self):
        pet = Pet(name='doggo').save()
        owner = Person(name='John').save()
        owner.proxy.pets.set([pet])

        pet.proxy.owner.fill()
        owner.proxy.pets.fill()

        self.assertIsNotNone(pet.owner)
        self.assertEqual(pet.owner.id, owner.id)

        self.assertTrue(type(owner.pets) == list)
        self.assertEqual(len(owner.pets), 1)
        self.assertEqual(owner.pets[0].id, pet.id)

    def test_delete_relation(self):
        c1 = Car().save()
        c2 = Car().save()

        d1 = Driver().save()

        d1.proxy.cars.set([c1, c2])

        self.assertTrue(c1 in d1.proxy.cars)
        self.assertTrue(c2 in d1.proxy.cars)
        self.assertTrue(d1 in c1.proxy.drivers)
        self.assertTrue(d1 in c2.proxy.drivers)

        d1.proxy.cars.remove(c1)

        self.assertFalse(c1 in d1.proxy.cars)
        self.assertTrue(c2 in d1.proxy.cars)
        self.assertFalse(d1 in c1.proxy.drivers)
        self.assertTrue(d1 in c2.proxy.drivers)

    def test_get_relation(self):
        c1 = Car().save()
        c2 = Car().save()

        d1 = Driver().save()

        d1.proxy.cars.set([c1, c2])

        cars = d1.proxy.cars.get()

        orig = sorted([c1, c2], key=lambda c:c.id)
        cars = sorted(cars, key=lambda c:c.id)

        self.assertListEqual(orig, cars)

    def test_get_foreignid_relation(self):
        pet = Pet().save()
        owner = Person().save()
        owner.proxy.pets.add(pet)

        self.assertEqual(pet.proxy.owner.get().id, owner.id)

        owner.proxy.pets.remove(pet)

        self.assertIsNone(pet.proxy.owner.get())


if __name__ == '__main__':
    unittest.main()



EPSILON = 0.00001

def debyte_string(byte_string):
    if type(byte_string) == bytes:
        return byte_string.decode('utf8')

    return byte_string

def debyte_iterator(iterator):
    return map(debyte_string, iterator)

def debyte_set(iterator):
    return set(debyte_iterator(iterator))

def debyte_list(iterator):
    return list(debyte_iterator(iterator))

def debyte_tuple(iterator):
    return tuple(debyte_iterator(iterator))

def debyte_float(byte_float):
    return float(byte_float)

def debyte_hash(dictlike):
    return dict(map(
        debyte_tuple,
        dictlike.items()
    ))


class Location:

    def __init__(self, lon, lat):
        self.lon = lon
        self.lat = lat

    def to_json(self):
        return {
            'lat': self.lat,
            'lon': self.lon,
        }

    def __eq__(self, other):
        if type(other) != Location:
            return False

        return abs(self.lat - other.lat)<EPSILON and abs(self.lon - other.lon)<EPSILON

    def __str__(self):
        return '<Location lat={} lon={}>'.format(self.lat, self.lon)



import unittest
from coralillo.utils import parse_embed


class UtilsTestCase(unittest.TestCase):

    def test_parse_embed(self):
        array = ['object']
        output = [['object', None]]
        self.assertListEqual(parse_embed(array), output)

        array = ['object.field']
        output = [['object', ['field']]]
        self.assertListEqual(parse_embed(array), output)

        array = ['object.field', 'foo', 'object.var']
        output = [['object', ['field', 'var']], ['foo', None]]
        self.assertListEqual(parse_embed(array), output)



from .hashing import make_password, is_hashed
from .errors import MissingFieldError, InvalidFieldError, ReservedFieldError, NotUniqueFieldError, DeleteRestrictedError
from .datamodel import debyte_string, debyte_hash
from . import datamodel
from importlib import import_module
from .utils import to_pipeline
import re
import datetime
import json


class Field:
    ''' Defines a field of a model. Represents how to store this specific
    datatype in the redis database '''

    def __init__(self, *, name=None, index=False, required=True, default=None, private=False, regex=None, forbidden=None, allowed=None, fillable=True):
        # This field's value is mapped to the ID in a redis hash so you can Model.get_by(field, value)
        self.index     = index     

        # This field is required in validation
        self.required  = required  

        # This field's default value
        self.default   = default   

        # This field's value is not published in the JSON representation of the object
        self.private   = private   

        # A regular expresion that validates this field's value
        self.regex     = regex     

        # A set of forbidden values for this field
        self.forbidden = forbidden

        # The set of only allowed valies for this field
        self.allowed = allowed

        # This field can't be set via http
        self.fillable  = fillable  

        # This will be set later by the proxy
        self.name      = name
        self.obj       = None

    def value_or_default(self, value):
        ''' Returns the given value or the specified default value for this
        field '''
        if value is None:
            if callable(self.default):
                return self.default()
            else:
                return self.default

        return value

    def validate_required(self, value):
        ''' Validates the given value agains this field's 'required' property
        '''
        if self.required and (value is None or value==''):
            raise MissingFieldError(self.name)

    def init(self, value):
        ''' Returns the value that will be set in the model when it is passed
        as an __init__ attribute '''
        return self.value_or_default(value)

    def recover(self, data, redis=None):
        ''' Retrieve this field's value from the database '''
        value = data.get(self.name)

        if value is None or value == 'None':
            return None

        return str(value)

    def prepare(self, value):
        ''' Prepare this field's value to insert in database '''
        if value is None: return None

        return str(value)

    def to_json(self, value):
        ''' Format the value to be presented in json format '''
        return value

    def save(self, value, redis, *, commit=True):
        ''' Sets this fields value in the databse '''
        value = self.prepare(value)

        if value is not None:
            redis.hset(self.obj.key(), self.name, value)
        else:
            redis.hdel(self.obj.key(), self.name)

        if self.index:
            key = self.key()

            if self.name in self.obj._old:
                redis.hdel(key, self.obj._old[self.name])

            redis.hset(key, value, self.obj.id)

    def delete(self, redis):
        ''' Deletes this field's value from the databse. Should be implemented
        in special cases '''
        if self.index:
            redis.hdel(self.key(), getattr(self.obj, self.name))

    def validate(self, value, redis):
        '''
        Validates data obtained from a request and returns it in the apropiate
        format
        '''
        # cleanup
        if type(value) == str:
            value = value.strip()

        value = self.value_or_default(value)

        # validation
        self.validate_required(value)

        if self.regex and not re.match(self.regex, value, flags=re.ASCII):
            raise InvalidFieldError(self.name)

        if self.forbidden and value in self.forbidden:
            raise ReservedFieldError(self.name)

        if self.allowed and value not in self.allowed:
            raise InvalidFieldError(self.name)

        if self.index:
            key = self.key()

            old = debyte_string(redis.hget(key, value))
            old_value = getattr(self.obj, self.name)

            if old is not None and old != self.obj.id:
                raise NotUniqueFieldError(self.name)
            elif old_value != value:
                self.obj._old[self.name] = old_value

        return value

    def key(self):
        return self.obj.cls_key() + ':index_' + self.name


class Text(Field):
    pass


class TreeIndex(Field):

    def save(self, value, redis, *, commit=True):
        ''' Sets this fields value in the databse '''
        value = self.prepare(value)

        if value is not None:
            redis.hset(self.obj.key(), self.name, value)
        else:
            redis.hdel(self.obj.key(), self.name)

        key = self.key()

        if self.name in self.obj._old:
            redis.hdel(key, self.obj._old[self.name])

        redis.sadd(key + ':' + value, self.obj.id)

    def delete(self, redis):
        ''' Deletes this field's value from the databse. Should be implemented
        in special cases '''
        value = getattr(self.obj, self.name)
        redis.srem(self.key() + ':' + value, self.obj.id)

    def key(self):
        return self.obj.cls_key() + ':tree_' + self.name


class Hash(Text):
    ''' A value that should be stored as a hash, for example a password '''

    def init(self, value):
        ''' hash passwords given in the constructor '''
        value = self.value_or_default(value)

        if value is None: return None

        if is_hashed(value):
            return value

        return make_password(value)

    def prepare(self, value):
        ''' Prepare this field's value to insert in database '''
        if value is None:
            return None

        if is_hashed(value):
            return value

        return make_password(value)

    def validate(self, value, redis):
        ''' hash passwords given via http '''
        value = super().validate(value, redis)

        if is_hashed(value):
            return value

        return make_password(value)


class Bool(Field):
    ''' A boolean value '''

    def validate(self, value, redis):
        value = self.value_or_default(value)

        if value is None: return None

        if type(value) == bool:
            return value

        return value == 'true' or value == '1'

    def prepare(self, value):
        return str(value)

    def recover(self, data, redis=None):
        value = data.get(self.name)

        if value is None or value == 'None':
            return None

        return value in ['True', 'true', '1', 1]


class Integer(Field):
    ''' An integer value '''

    def validate(self, value, redis):
        value = self.value_or_default(value)

        self.validate_required(value)

        if value is None:
            return None

        try:
            return int(value)
        except ValueError:
            raise InvalidFieldError(self.name)

    def recover(self, data, redis=None):
        value = data.get(self.name)

        if value == '' or value is None or value == 'None':
            return None

        return int(value)

    def prepare(self, value):
        return str(value)


class Float(Field):

    def validate(self, value, redis):
        value = self.value_or_default(value)

        self.validate_required(value)

        try:
            return float(value)
        except ValueError:
            raise InvalidFieldError(self.name)

    def recover(self, data, redis=None):
        value = data.get(self.name)

        if value == '' or value is None or value == 'None':
            return None

        return float(value)

    def prepare(self, value):
        return str(value)


class Datetime(Field):
    ''' A datetime that can be used transparently as such in the model '''

    def validate(self, value, redis):
        '''
        Validates data obtained from a request in ISO 8061 and returns it in Datetime data type
        '''

        value = self.value_or_default(value)

        self.validate_required(value)

        if value is None:
            return None

        if type(value) == str:
            try:
                value = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%SZ')
            except ValueError:
                raise InvalidFieldError(self.name)

        return value

    def recover(self, data, redis=None):
        value = data.get(self.name)

        if not value or value == 'None':
            return None

        return datetime.datetime.utcfromtimestamp(int(value))

    def prepare(self, value):
        if value is None:
            return None

        return str(int(value.timestamp()))

    def to_json(self, value):
        if value is None:
            return None

        return value.replace(microsecond=0).isoformat() + 'Z'


class Location(Field):
    ''' A geolocation '''

    def prepare(self, value):
        return value

    def save(self, value, redis, *, commit=True):
        key = self.key()

        if value is not None:
            redis.geoadd(key, value.lon, value.lat, self.obj.id)
        else:
            redis.zrem(key, self.obj.id)

    def delete(self, redis):
        key = self.key()

        redis.zrem(key, self.obj.id)

    def to_json(self, value):
        if value is None:
            return None

        return value.to_json()

    def recover(self, data, redis):
        key = self.key()

        try: # TODO change this once the GEO api is stable in redis-py
            value = redis.geopos(key, self.obj.id)
        except TypeError:
            value = None

        if not value:
            return None

        if value[0] is None:
            return None

        return datamodel.Location(*value[0])

    def validate(self, value, redis):
        value = self.value_or_default(value)

        self.validate_required(value)

        if value is None:
            return None

        try:
            lon, lat = map(float, value.split(','))

            assert -180 < lon < 180
            assert -90 < lat < 90

            return datamodel.Location(lon, lat)
        except:
            raise InvalidFieldError(self.name)

    def key(self):
        return self.obj.cls_key() + ':geo_' + self.name


class Dict(Field):
    ''' A dict that can be used transparently as such in the model '''

    def validate(self, value, redis):
        if not value:
            return dict()

        try:
            value = json.loads(value)
        except json.decoder.JSONDecodeError as e:
            raise InvalidFieldError(self.name)

        return value

    def prepare(self, value):
        return value

    def save(self, value, redis, *, commit=True):
        key = self.key()

        if bool(value) is not False:
            redis.delete(key)
            redis.hmset(key, value)
        else:
            redis.delete(key)

    def delete(self, redis):
        key = self.key()

        redis.delete(key)

    def to_json(self, value):
        if value is None:
            return dict()

        return value

    def recover(self, data, redis):
        key = self.key()

        try:
            value = debyte_hash(redis.hgetall(key))
        except TypeError:
            value = dict()

        return value

    def key(self):
        return '{}:{}:dict_{}'.format(self.obj.cls_key(), self.obj.id, self.name)


class Relation(Field):

    def __init__(self, model, *, private=False, on_delete=None, inverse=None):
        self.modelspec = model
        self.private   = private
        self.on_delete = on_delete
        self.inverse   = inverse
        self.fillable  = False

    def model(self):
        if type(self.modelspec) == str:
            from . import Model

            pieces = self.modelspec.split('.')

            return getattr(import_module('.'.join(pieces[:-1])), pieces[-1])

        return self.modelspec


class ForeignIdRelation(Relation):

    def __init__(self, model, *, private=False, on_delete=None, inverse=None):
        super().__init__(model, private=private, on_delete=on_delete, inverse=inverse)
        self.default   = None

    def relate(self, obj, pipeline):
        pipeline.hset(self.obj.key(), self.name, obj.id)

    def unrelate(self, obj, redis):
        redis.hdel(self.obj.key(), self.name, obj.id)

    def set(self, value, *, commit=True):
        redis = type(self.obj).get_redis()
        getattr(self.obj.proxy, self.name).fill()

        prev = getattr(self.obj, self.name)

        if prev is not None:
            getattr(prev.proxy, self.inverse).unrelate(self.obj, redis)

        if value is None:
            redis.hdel(self.obj.key(), self.name)
            setattr(self.obj, self.name, None)
            return

        redis.hset(self.obj.key(), self.name, value.id)

        related = value

        if self.inverse:
            getattr(related.proxy, self.inverse).relate(self.obj, redis)

        setattr(self.obj, self.name, value)

    def delete(self, redis):
        getattr(self.obj.proxy, self.name).fill()
        item = getattr(self.obj, self.name)

        if item is None: return

        if self.on_delete == 'restrict':
            raise DeleteRestrictedError('attempt to delete with relations and restrict flag')

        if self.on_delete == 'cascade':
            item.delete()
        elif self.inverse:
            getattr(item.proxy, self.inverse).unrelate(self.obj, redis)

    def fill(self):
        setattr(self.obj, self.name, self.get())

    def get(self):
        redis = type(self.obj).get_redis()
        value = debyte_string(redis.hget(self.obj.key(), self.name))

        return self.model().get(value)


class MultipleRelation(Relation):

    def relate_all(self, value, pipe):
        raise NotImplementedError('must be implemented in subclass')

    def set(self, value, *, commit=True):
        key  = self.key()
        redis = type(self.obj).get_redis()
        pipe = to_pipeline(redis)

        pipe.delete(key)

        if type(value) != list or len(value) == 0:
            setattr(self.obj, self.name, [])
            return

        self.relate_all(value, pipe)

        for related in value:
            if self.inverse:
                getattr(related.proxy, self.inverse).relate(self.obj, pipe)

        if commit:
            pipe.execute()

        setattr(self.obj, self.name, value)

    def add(self, value):
        redis = type(self.obj).get_redis()

        self.relate(value, redis)

        if self.inverse:
            getattr(value.proxy, self.inverse).relate(self.obj, redis)

    def recover(self, data, redis=None):
        ''' Don't read the database by default '''
        return []

    def prepare(self, value):
        return value

    def init(self,value):
        proposed = self.value_or_default(value)

        if not proposed:
            return []
        return proposed

    def fill(self, **kwargs):
        ''' Loads the relationships into this model. They are not loaded by
        default '''
        setattr(self.obj, self.name, self.get(**kwargs))

    def get(self, **kwargs):
        ''' Returns this relation '''
        redis = type(self.obj).get_redis()
        related = list(map(
            lambda id : self.model().get(debyte_string(id)),
            self.get_related_ids(redis, **kwargs)
        ))

        return related

    def delete(self, redis):
        key = self.key()

        getattr(self.obj.proxy, self.name).fill()
        items = getattr(self.obj, self.name)

        if self.on_delete == 'restrict' and len(items) > 0:
            raise DeleteRestrictedError('attempt to delete with relations and restrict flag')

        for item in items:
            if self.on_delete == 'cascade':
                item.delete()
            elif self.inverse:
                getattr(item.proxy, self.inverse).unrelate(self.obj, redis)

        redis.delete(key)

    def remove(self, value):
        redis = type(self.obj).get_redis()

        self.unrelate(value, redis)

        if self.inverse:
            getattr(value.proxy, self.inverse).unrelate(self.obj, redis)

    def count(self):
        raise NotImplementedError('count is not implemented for this subclass of MultipleRelation')


class SetRelation(MultipleRelation):
    ''' A relationship with another model '''

    def __init__(self, model, *, private=False, on_delete=None, inverse=None):
        super().__init__(model, private=private, on_delete=on_delete, inverse=inverse)
        self.default   = []
        self.fillable  = False

    def key(self):
        return '{}:{}:srel_{}'.format(self.obj.cls_key(), self.obj.id, self.name)

    def relate(self, obj, redis):
        redis.sadd(self.key(), obj.id)

    def relate_all(self, value, redis):
        redis.sadd(self.key(), *[r.id for r in value])

    def unrelate(self, obj, redis):
        redis.srem(self.key(), obj.id)

    def get_related_ids(self, redis):
        key = self.key()

        return redis.smembers(key)

    def count(self):
        key   = self.key()
        redis = type(self.obj).get_redis()

        return redis.scard(key)

    def __contains__(self, item):
        if not isinstance(item, self.model()):
            return False

        redis = type(self.obj).get_redis()

        return redis.sismember(self.key(), item.id)


class SortedSetRelation(MultipleRelation):
    ''' A relationship with another model that ensures the same ordering every
    time '''

    def __init__(self, model, sort_key, **kwargs):
        super().__init__(model, **kwargs)
        self.sort_key = sort_key
        self.default   = []
        self.fillable  = False

    def key(self):
        return '{}:{}:zrel_{}'.format(self.obj.cls_key(), self.obj.id, self.name)

    def relate(self, obj, redis):
        field = getattr(self.model(), self.sort_key) # the field in the foreign model

        redis.zadd(self.key(), field.prepare(getattr(obj, self.sort_key)), obj.id)

    def relate_all(self, value, redis):
        field = getattr(self.model(), self.sort_key) # the field in the foreign model

        p = lambda v: int(field.prepare(getattr(v, self.sort_key)))

        redis.zadd(self.key(), *sum(([p(r), r.id] for r in value), []))

    def unrelate(self, obj, redis):
        redis.zrem(self.key(), obj.id)

    def get_related_ids(self, redis, *, score=None):
        key = self.key()

        if score:
            return redis.zrangebyscore(key, *score)

        return redis.zrange(key, 0, -1)

    def count(self):
        key   = self.key()
        redis = type(self.obj).get_redis()

        return redis.zcard(key)

    def __contains__(self, item):
        if not isinstance(item, self.model()):
            return False

        redis = type(self.obj).get_redis()

        return redis.zscore(self.key(), item.id) is not None



from coralillo import Engine, Model, fields
from coralillo.errors import UnboundModelError
from random import choice
import unittest


class EngineTestCase(unittest.TestCase):

    def test_create_engine_with_url(self):
        url = 'redis://localhost:6379/0'

    def test_create_engine(self):
        eng1 = Engine(db=11)
        eng2 = Engine(db=12)

        class Dog(Model):
            name = fields.Text()

            class Meta:
                engine = eng1

        class Cat(Model):
            name = fields.Text()

            class Meta:
                engine = eng2

        doggo = Dog(name='doggo').save()
        catto = Cat(name='catto').save()

        self.assertTrue(eng1.redis.exists('dog:{}:obj'.format(doggo.id)))
        self.assertFalse(eng1.redis.exists('cat:{}:obj'.format(catto.id)))

        self.assertTrue(eng2.redis.exists('cat:{}:obj'.format(catto.id)))
        self.assertFalse(eng2.redis.exists('dog:{}:obj'.format(doggo.id)))

    def test_set_engine_delayed(self):
        class Foo(Model):
            name = fields.Text()

        with self.assertRaises(UnboundModelError):
            Foo.get_engine()

        eng = Engine()

        Foo.set_engine(eng)

        self.assertIsNone(Foo.get('de')) # Only possible if bound to engine

    def test_unbound_models(self):
        class Dog(Model):
            name = fields.Text()

        with self.assertRaises(UnboundModelError):
            Dog(mame='doggo').save()

    def test_register_lua_script(self):
        eng = Engine(db=0)

        eng.lua.register('my_script', 'return ARGV[1]')

        self.assertIsNotNone(eng.lua.my_script)
        self.assertEqual(eng.lua.my_script(args=[4]), b'4')

    def test_can_replace_id_function(self):
        def simple_ids():
            return ''.join(choice('123456789abcdef') for c in range(11))

        simple_eng = Engine(id_function=simple_ids)
        uuid_eng = Engine()

        class SimpleDog(Model):
            name = fields.Text()

            class Meta:
                engine = simple_eng

        class UuidDog(Model):
            name = fields.Text()

            class Meta:
                engine = uuid_eng

        simple_doggo = SimpleDog(name='doggo').save()
        self.assertEqual(len(simple_doggo.id), 11)

        uuid_doggo = UuidDog(name='doggo').save()
        self.assertEqual(len(uuid_doggo.id), 32)


if __name__ == '__main__':
    unittest.main()



import os

SCRIPT_PATH = os.path.dirname(__file__)

class Lua:

    def __init__(self, redis):
        self.redis = redis
        scripts = filter(lambda s:s.endswith('.lua'), os.listdir(SCRIPT_PATH))

        for scriptname in scripts:
            with open(os.path.join(SCRIPT_PATH, scriptname)) as script:
                setattr(self, scriptname.split('.')[0], redis.register_script(script.read()))

    def register(self, name, contents):
        setattr(self, name, self.redis.register_script(contents))



import json


class ImproperlyConfiguredError(Exception): pass

class ModelNotFoundError(Exception): pass

class DeleteRestrictedError(Exception): pass

class UnboundModelError(Exception): pass


class ValidationErrors(Exception):

    def __init__(self):
        self.errors = []

    def append(self, e):
        self.errors.append(e)

    def has_errors(self):
        return len(self.errors)

    def to_json(self):
        return [e.to_json() for e in self.errors]


class BadField(Exception):

    message = '{field} is invalid'
    errorcode = 'invalid'

    def __init__(self, field='', value=''):
        assert type(field) == str, 'Field must be string'

        self.field = field
        self.value = value

    def get_detail(self):
        return self.message.format(
            field = self.field,
            value = self.value,
        )

    def to_json(self):
        return {
            'detail': self.get_detail(),
            'field': self.field,
            'i18n': 'errors.{field}.{error}'.format(
                field = self.field,
                error = self.errorcode,
            ),
        }

class MissingFieldError(BadField):
    message = '{field} is required'
    errorcode = 'required'

class InvalidFieldError(BadField):
    message = '{field} is not valid'
    errorcode = 'invalid'

class ReservedFieldError(BadField):
    message = '{field} is reserved'
    errorcode = 'reserved'

class NotUniqueFieldError(BadField):
    message = '{field} is not unique'
    errorcode = 'unique'
